package org.vishia.fpga.exmpl2;

import org.vishia.fpga.Fpga;
import org.vishia.fpga.FpgaModule_ifc;
import org.vishia.fpga.stdmodules.Bit_ifc;
import org.vishia.fpga.stdmodules.Reset_ifc;
import org.vishia.fpga.stdmodules.Word_ifc;

public final class MdlA implements FpgaModule_ifc, SpecB_ifc, Word_ifc {
  protected class Ref {
    final Reset_ifc res;
    final Bit_ifc inpX;
    final Bit_ifc refC;
    Ref ( Reset_ifc res, Bit_ifc inpX, Bit_ifc refC) {
      this.res = res;
      this.inpX = inpX;
      this.refC = refC;
    }
  }
  
  protected Ref ref;
  
  public MdlA ( Reset_ifc res, Bit_ifc inpX, Bit_ifc refC) {
    this.ref = new Ref(res, inpX, refC);
  }
  
  public MdlA ( ) {}  // init() necessary to use
  
  public void init ( Reset_ifc res, Bit_ifc inpX, Bit_ifc refC) {
    this.ref = new Ref(res, inpX, refC);
  }

  @Fpga.VHDL_PROCESS public static final class Q {
    @Fpga.STDVECTOR(4) final int ct;
    final boolean ce;
    
    //...
    Q() {                              // produce hardware reset state
      this.ct = 0;
      this.ce = false;
    }
    
    @Fpga.VHDL_PROCESS Q(int time, Q z, Ref ref) {
      this.ct = z.ct -1;
      this.ce = z.ct == 0b0000;
    }
  }
  
  private Q q_d = new Q(); 
  protected Q q;

  @Fpga.VHDL_PROCESS public static final class Process2 {
    final char s1;
    @Fpga.BITVECTOR(7) final int sh;
    @Fpga.STDVECTOR(20) final int ct;
    //...
    Process2(){
      this.s1 = '0';
      this.sh = 0;
      this.ct = 0;
    }

    @Fpga.VHDL_PROCESS Process2(int time, Process2 z, Ref ref, MdlA thism) {
      if(thism.q.ce) {
        this.s1 = z.s1;
        this.sh = z.sh;
        this.ct = z.ct +1;
      } else {
        this.s1 = z.s1;
        this.sh = z.sh;
        this.ct = z.ct;
      }
    }
  }

  private Process2 process2_d = new Process2(); 
  protected Process2 process2;

  
  @Override public void reset () {
    this.q = this.q_d = new Q();
    this.process2 = this.process2_d = new Process2();
  }

  @Override public void step ( int time ) {
    this.q_d = new Q(time, this.q, this.ref);
    this.process2_d = new Process2(time, this.process2, this.ref, this);
  }

  @Override public void update () {
    this.q = this.q_d;
    this.process2 = this.process2_d;
  }

  @Override public int getWord ( int time, int min ) {
    return this.q.ct;
  }

  
}
